@@ -1,5 +1,5 @@
 package Git::Raw;
-$Git::Raw::VERSION = '0.65';
+$Git::Raw::VERSION = '0.66';
 use strict;
 use warnings;
 
@@ -11,11 +11,16 @@ use Git::Raw::Error::Category;
 use Git::Raw::Packbuilder;
 use Git::Raw::Blob;
 use Git::Raw::Commit;
+use Git::Raw::Index;
+use Git::Raw::Indexer;
+use Git::Raw::Odb;
+use Git::Raw::Mempack;
 use Git::Raw::Reference;
 use Git::Raw::Repository;
 use Git::Raw::Stash;
 use Git::Raw::Stash::Progress;
 use Git::Raw::Tree;
+use Git::Raw::TransferProgress;
 
 =for HTML
 <a href="https://travis-ci.org/jacquesg/p5-Git-Raw">
@@ -35,7 +40,7 @@ Git::Raw - Perl bindings to the Git linkable library (libgit2)
 
 =head1 VERSION
 
-version 0.65
+version 0.66
 
 =head1 DESCRIPTION
 
