@@ -1,5 +1,5 @@
 package Git::Raw::Diff;
-$Git::Raw::Diff::VERSION = '0.65';
+$Git::Raw::Diff::VERSION = '0.66';
 use strict;
 use warnings;
 
@@ -11,7 +11,7 @@ Git::Raw::Diff - Git diff class
 
 =head1 VERSION
 
-version 0.65
+version 0.66
 
 =head1 DESCRIPTION
 
@@ -22,10 +22,19 @@ B<WARNING>: The API of this module is unstable and may change without warning
 
 =head1 METHODS
 
+=head2 new( $buffer )
+
+Create a diff from the content of a git patch.
+
 =head2 merge( $from )
 
 Merge the given diff with the L<Git::Raw::Diff> C<$from>.
 
+=head2 buffer( $format )
+
+Get the content of a diff as text. See C<Git::Raw::Diff-E<gt>print()> for valid
+C<$format> values.
+
 =head2 delta_count( )
 
 Query how many diff records there are in the diff.
