@@ -1,11 +1,11 @@
 package Git::Raw::Error;
-$Git::Raw::Error::VERSION = '0.65';
+$Git::Raw::Error::VERSION = '0.66';
 use strict;
 use warnings;
 use Carp;
 
 use overload
-	'""'       => sub { return $_[0] -> message },
+	'""'       => sub { return $_[0] -> message.' at '.$_[0] -> file.' line '.$_[0] -> line },
 	'0+'       => sub { return $_[0] -> code },
 	'bool'     => sub { return $_[0] -> _is_error },
 	'fallback' => 1;
@@ -35,7 +35,7 @@ Git::Raw::Error - Error class
 
 =head1 VERSION
 
-version 0.65
+version 0.66
 
 =head1 DESCRIPTION
 
@@ -51,6 +51,14 @@ B<WARNING>: The API of this module is unstable and may change without warning
 
 Error message.
 
+=head2 file( )
+
+Caller file.
+
+=head2 line( )
+
+Caller line.
+
 =head2 code( )
 
 Error code.
