@@ -1,5 +1,5 @@
 package Git::Raw::Repository;
-$Git::Raw::Repository::VERSION = '0.65';
+$Git::Raw::Repository::VERSION = '0.66';
 use strict;
 use warnings;
 
@@ -11,7 +11,7 @@ Git::Raw::Repository - Git repository class
 
 =head1 VERSION
 
-version 0.65
+version 0.66
 
 =head1 SYNOPSIS
 
@@ -116,7 +116,14 @@ Retrieve the default L<Git::Raw::Config> of the repository.
 =head2 index( [$new_index] )
 
 Retrieve the index of the repository. If C<$new_index> is passed, it will be used
-as the index of the repository. Returns a L<Git::Raw::Index> object.
+as the index of the repository. If C<$new_index> is C<undef> the index associated
+with the repository will be disassociated. Returns a L<Git::Raw::Index> object or
+undef if index has been disassociated as part of the call.
+
+=head2 odb( [$new_odb] )
+
+Retrieve the object database of the repository. If C<$odb> is passed, it will be
+used as the object database. Returns a L<Git::Raw::Odb> object.
 
 =head2 head( [$new_head] )
 
